#ifndef _MUSSA_FLP_H_
#define _MUSSA_FLP_H_
//  This file is part of the Mussa source distribution.
//  http://mussa.caltech.edu/
//  Contact author: Tristan  De Buysscher, tristan@caltech.edu

// This program and all associated source code files are Copyright (C) 2005
// the California Institute of Technology, Pasadena, CA, 91125 USA.  It is
// under the GNU Public License; please see the included LICENSE.txt
// file for more information, or contact Tristan directly.


//                        ----------------------------------------
//                            ---------- flp.hh  -----------
//                        ----------------------------------------
#include <boost/filesystem/path.hpp>

#include <list>
#include <string>
#include <vector>
#include <iostream>

class Sequence;
//! FLP = Fixed Length Pairs (Data)
/*!
 * vector of linked lists of the match type struct
 */
class FLPs
{
public:
    typedef size_t size_type;
    
    FLPs();
    FLPs(const FLPs& );
    //! Setup a FLP and reserve space for the match lists
    /*!
     * Initialize the all_matches structure with a list of matches
     * for enough windows for the size of sequence 1
     */
    
    struct match
    {
      int index;
      int score;

      friend bool operator==(const match&, const match& );
      friend std::ostream &operator<<(std::ostream&, const match&);
    };

    void setup(int win_size, int hard_thres);
    //! compare two sequences and store the list of results in all_matches
    void seqcomp(const Sequence& seq1, const Sequence& seq2, bool is_RC);
    //bool FLPs::match_less(match *match1, match *match2);
    //void FLPs::sort();
    //! Return all the matches for a particular window? (index and score pairs)
    std::list<match> matches(int index) const;
    //! Return all the match indexes for a particular window?
    std::list<int> match_locations(int index) const;
    //! Return all the match indexes for a particular window above a threshold
    std::list<int> thres_matches(int index, int thres) const;
    //! Return the number of windows stored in this FLPs
    /*! (this is should be the same as seq1_len-window_size+1
     */
    int size() const;
    //! Save all the FLPs to save_file_path
    void save(boost::filesystem::path save_file_path);
    //! Load a vector of a lists of FLPs
    void load(boost::filesystem::path file_path);

    //! how far we are through current seqcomp
    float progress() const;

    const static int seqcomp_not_running = -1;
  private:
    //! the number of base pairs in the sliding window
    int window_size;
    //! the minimum tnumber of base pairs need for this window to be saved.
    int hard_threshold;

    //! add a match from location seq1_i to seq2_i with a threshold of a_score
    /*! i2_offset is used to shift the window start when doing a reverse 
     * compliment. (This is called by seqcomp() )
     */
    inline void add(int seq1_i, int seq2_i, int a_score, int i2_offset);
    //! this list of matches between the two sequences
    /*! All of the matches are stored here, it appears each window position
     *  in sequence 1 gets an entry in the vector, the list contains matches
     *  between sequence 1 and sequence 2
     */
    std::vector<std::list<match> > *all_matches;
    //! reserve space for the appropriate number of windows.
    /*! this is mostly so seqcomp can use operator[]
     */
    void alloc_matches(std::string::size_type len1=0);
    
    //! make sure that a sequence is acceptable to seqcomp
    void validate_sequence(const Sequence&) const;
    
    //! current loop index
    int seqcomp_i;
    //! end seqcomp index (when terminating, seqcomp_i == seqcomp_end.
    //! when not running seqcomp_i == seqcomp_end == seqcomp_not_running
    int seqcomp_end;
};
#endif
